#
# Copyright 2020 Thomas Bastian, Jeffrey Goff, Albert Pang
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
#
# ContentProvider
#
# Provides some content through getContent()
#
# Authors: Thomas Bastian, Albert Pang
#
'''

import json
import logging
import os

import requests
import urllib3
from yaml import dump

from .RedisStore import RedisStore
from .util import convertToDict

DEFAULT_API_TIMEOUT = 15  # Default timeout value (s) for APIContentProvider

urllib3.disable_warnings()

# Logger
log = logging.getLogger(__name__)


class ContentProvider():
    def __init__(self):
        pass

    def __str__(self):
        return ""

    def getContent(self):
        return None

    def getContentData(self):
        ''' Attempt to decode the content read and returns a dictionary or a list.
        Will attempt to decode the content as JSON or YAML

        Parameter
        ---------
            None

        Returns
        -------
            dict or list : A dictionary if able to convert content as dict or list. None if unsuccessful
        '''
        content = self.getContent()

        return convertToDict(content)

    def reprJSON(self):
        data = {}
        data['ContentProvider'] = str(self)
        return data

    def getContentJSON(self):
        '''Attempt to return a JSON object from the content source.

        Parameters
        ----------
            None

        Returns
        -------
            str : A JSON object string. Returns None if unable to convert content to JSON object
        '''
        data = self.getContentData()

        if isinstance(data, dict) or isinstance(data, list):
            log.debug("Found dict or list")
            return json.dumps(data)
        else:
            return None

    def getContentYml(self):
        '''Attempt to return a YAML formatted string from the content source.

        Parameters
        ----------
            None

        Returns
        -------
            str : A string in YAML format. Returns None if unable to convert content to JSON object
        '''
        data = self.getContentData()

        if isinstance(data, dict) or isinstance(data, list):
            log.debug("Found dict or list")
            return dump(data)
        else:
            return None

    def __iter__(self):
        return None

    def _next(self):
        raise StopIteration

    def __next__(self):
        return self._next()

    def next(self):
        return self._next()


class FileContentProvider(ContentProvider):
    def __init__(self, filename):
        self.filename = filename

    def __str__(self):
        return "file://" + self.filename

    def getContent(self):
        fileHandle = open(self.filename, 'r')
        fileContent = fileHandle.read()
        fileHandle.close()
        return fileContent


class MultiFileContentProvider(ContentProvider):
    def __init__(self, dirname, endswith):
        self.dirname = dirname
        self.endswith = endswith
        self.fileIterator = None

    def __str__(self):
        return "dir://" + self.dirname + ", ending with: " + self.endswith

    def __repr__(self):
        return "dir://" + self.dirname + ", ending with: " + self.endswith

    def getContent(self):
        raise "unimplemented"

    def __iter__(self):
        self.fileIterator = iter(os.listdir(self.dirname))
        return self

    def _next(self):
        try:
            while True:
                filename = next(self.fileIterator)
                if filename.endswith(self.endswith):
                    break
            p = os.path.join(self.dirname, filename)
            return FileContentProvider(p)

        except StopIteration:
            raise StopIteration


class RedisStoreContentProvider(ContentProvider):
    def __init__(self, redisUrl, redisKey):
        self.redisUrl = redisUrl
        self.redisKey = redisKey

    def __str__(self):
        return self.redisUrl + "/" + self.redisKey

    def getContent(self):
        r = RedisStore(url=self.redisUrl)
        content = r.get(self.redisKey)
        return content


class MultiKeyRedisStoreContentProvider(ContentProvider):
    def __init__(self, redisUrl, redisKeysPatterns):
        self.redisUrl = redisUrl
        self.redisKeysPattern = redisKeysPatterns

    def __str__(self):
        return self.redisUrl + "/" + self.redisKeysPattern

    def __repr__(self):
        return self.redisUrl + "/" + self.redisKeysPattern

    def getContent(self):
        raise "unimplemented"

    def __iter__(self):
        r = RedisStore(url=self.redisUrl)
        redisKeys = r.keys(pattern=self.redisKeysPattern)
        self.fileIterator = iter(redisKeys)
        return self

    def _next(self):
        try:
            while True:
                redisKey = next(self.fileIterator)
                break
            return RedisStoreContentProvider(self.redisUrl, redisKey)

        except StopIteration:
            raise StopIteration


class ApiContentProvider(ContentProvider):
    def __init__(self, apiUrl, endpoint, params=None, sip_mode=True,
                 verify=True, timeout=DEFAULT_API_TIMEOUT):
        '''Retrieve content using REST API (http GET only)

        Parameters
        ----------
        apiUrl : str
            The host portion of the complete REST API call.
                E.g.
                    https://www.google.com
                    http://localhost:5000
            Trailing '/' is optional
        endpoint : str
            The endpoint of the complete REST API call
                E.g.
                    /endpoint
                    endpoint
            Leading '/' is optional
        params : dict, default=None
            list of tuples or bytes to send in the query string
            Do not set this if you want to use the 'getContentByEndpointParam()'
            method to add parameters dynamically for each call, or un
        sip_mode : bool, default=True
            Legacy behaviour for Aruba SIP. The message contained in JSON object
            is expected to be in the 'msg' attribute. e.g.
                { 'msg': {'a': 'value1', 'b': 'value2'} }
            The value of the 'msg' attribute (i.e. {'a': 'value1', 'b': 'value2'})
            instead of the whole thing.  Set to False to disable this behaviour.
            For backward compatibility in SIP software, this behaviour is enabled
            by default.
            TODO: refactor RR and ND scripts and change default to sip_mode=False
        verify : bool or str, default=True
            Either a boolean, in which case it controls whether we verify
            the server's TLS certificate, or a string, in which case it must be a path
            to a CA bundle to use.
        timeout : int, default=DEFAULT_API_TIMEOUT
        '''
        self.apiUrl = apiUrl.rstrip('/') if apiUrl else ""
        self.endpoint = endpoint if endpoint else ""
        self.params = params
        self.sip_mode = sip_mode
        self._verify = verify
        self._timeout = timeout
        if self.endpoint.startswith("/"):
            self._url = self.apiUrl + self.endpoint
        else:
            self._url = self.apiUrl + "/" + self.endpoint

    def __str__(self):
        return self._url

    def getContent(self):
        try:
            if self.params is not None:
                response = requests.request("GET", self._url,
                                            params=self.params, verify=self._verify,
                                            timeout=self._timeout)
            else:
                response = requests.request("GET", self._url,
                                            verify=self._verify,
                                            timeout=self._timeout)
            # NOTE: the 'msg' is for SIP backward-compatibility
            if self.sip_mode:
                return json.dumps(response.json()['msg'])
            else:
                return json.dumps(response.json())
        except Exception as e:
            log.error(f"Cannot retrieve content from {str(self)} ({e})")
            return ""

    def getContentByEndpointParam(self, param=None):
        '''
        Similar to getContent(), instead of passing the parameters in the form or URL query string
        such as:
            /url/endpoint?key1=value1&key2=value2
        pass the single parameter as a suffix of the URL

            /url/endpoint/param
        '''
        try:
            if param:
                if param.startswith("/"):
                    url = self._url + param
                else:
                    url = self._url + "/" + param
                response = requests.request("GET", url, verify=self._verify,
                                            timeout=self._timeout)
            else:
                response = requests.request("GET", self._url,
                                            verify=self._verify, timeout=self._timeout)
            # NOTE: the 'msg' is for SIP backward-compatibility
            if self.sip_mode:
                return json.dumps(response.json()['msg'])
            else:
                return json.dumps(response.json())
        except Exception as e:
            log.error(f"Cannot retrieve content from {str(self)} ({e})")
            return ""

    def getContentByEndpointParamData(self, param=None):
        ''' Similar to getContentByEndpointParam(), except attempt to convert
        retrieved content to dict
        '''

        content = self.getContentByEndpointParam(param=param)

        return convertToDict(content)



#
# Copyright 2021 Thomas Bastian, Jeffrey Goff, Albert Pang
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
#
# Generic CLI parser framework.
# - ARCLI is the only concrete implementation available.
#
# Authors: Thomas Bastian, Jeff Goff
#
'''

'''
# Parameters
'''
'''
#
'''

import logging
from .arcli.ArCliParser import ArCliParser
from .arcli.ArCliContext import ArCliContext


# Logger
log = logging.getLogger(__name__)


# Generic parser factory
class CliParserFactory():
    def __init__(self):
        pass

    @staticmethod
    def getParser(parserName, **parserKwArgs):
        arCliContext = ArCliContext()
        arCliContext.loadArCliDocuments(parserKwArgs['arCliConfig'])
        arCliParser = ArCliParser(arCliContext)
        return arCliParser



#
# Copyright 2020 Thomas Bastian, Jeffrey Goff, Albert Pang
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
#
# Miscellaneous statistics classes and helpers.
#
# Authors: Thomas Bastian, Albert Pang
#
'''
import logging
import time
from datetime import datetime, timezone

from munch import DefaultMunch
'''
# Parameters
'''
INTERVAL_PRECISION = 2  # output should be rounded to this number of significant digits
RATE_PRECISION = 3
UNKNOWN_LABEL = "UNKNOWN"

# Logger
log = logging.getLogger(__name__)


def timestamp2str(ts):
    utc_time = datetime.fromtimestamp(ts, timezone.utc)
    local_time = utc_time.astimezone()
    return local_time.strftime("%Y-%m-%d %H:%M:%S.%f%z (%Z)")


class Counter():
    '''
    DEPRECATED: use CounterEvent instead
    This class is for counting metrics. Each metric has a topic and a group.
    Count and rates for each metric are calculated for each individual topic as well as at the group level.
    Percentage of each topic within a group is also calculated.
    '''
    def __init__(self, name="undefined", start_ts=time.time()):
        self._name = name
        self._start_ts = start_ts
        self._stop_ts = None

        self._topic_counts = {}
        self._group_counts = {}

        self._topic_group_mapping = {}  # Keep tracks of which group a stat topic belongs to

        self._last_calculated_rates = None

    def increment(self, topic, group="default", increment=1, timestamp=None):
        log.debug(f"topic: {topic} group: {group} increment: {increment}")
        if not isinstance(topic, (str, int, float, bool)):
            log.error(f"invalid type for key ({topic} to increment)")
            return None
        if not isinstance(group, (str, int, float, bool)):
            log.error(f"invalid type for group ({group} to increment)")
            return None
        if group not in self.groups:
            self._topic_counts[group] = {}
        if topic in self.topics(group):
            self._topic_counts[group][topic] += increment
        else:
            self._topic_counts[group][topic] = increment

        if group in self.groups:
            self._group_counts[group] += increment
        else:
            self._group_counts[group] = increment

    def rates(self, stop_ts=time.time(), interval=None):
        '''
        calculates the rates of the counters based on the counts
        from start_ts to stop_ts.  If interval is defined, use the
        value (in seconds) as the duration of the counter rather
        than start_ts and stop_ts
        '''
        self._stop_ts = stop_ts

        if interval:
            interval_for_calculation = interval
        else:
            interval_for_calculation = round(stop_ts - self._start_ts, INTERVAL_PRECISION)

        if interval_for_calculation <= 0:
            log.error(f"Cannot calculate rates for {self.name} counter as the interval is 0")
            return {}

        topic_rates = {}
        group_rates = {}
        topic_percentage = {}

        results = {}
        results['number_of_topics'] = {}

        for group in self.groups:
            results['number_of_topics'][group] = len(self.topics(group))
            for topic, count in self._topic_counts[group].items():
                if group not in topic_rates.keys():
                    topic_rates[group] = {}
                    topic_percentage[group] = {}
                topic_rates[group][topic] = round(count / interval_for_calculation, RATE_PRECISION)
                topic_percentage[group][topic] = round(count / self._group_counts[group], RATE_PRECISION)

        for group, count in self._group_counts.items():
            group_rates[group] = round(count / interval_for_calculation, RATE_PRECISION)

        results['interval'] = interval_for_calculation
        results['number_of_groups'] = len(self.groups)
        results['topic_rates'] = topic_rates
        results['group_rates'] = group_rates
        results['topic_percentage'] = topic_percentage

        return results

    def update_rates(self, stop_ts=time.time(), interval=None):
        self._last_calculated_rates = self.rates(stop_ts=stop_ts, interval=interval)

    def snapshot(self, stop_ts=None, interval=None, update_rates=False):
        if stop_ts is None:
            stop_ts = time.time()
        if update_rates:
            self.update_rates(stop_ts=stop_ts)

        self._stop_ts = stop_ts

        jsonData = {}
        jsonData['name'] = self.name
        jsonData['start_ts'] = self._start_ts
        jsonData['stop_ts'] = self._stop_ts
        jsonData['uptime'] = round(self._stop_ts - self._start_ts, INTERVAL_PRECISION)
        jsonData['start_time_str'] = timestamp2str(self._start_ts)
        jsonData['stop_time_str'] = timestamp2str(self._stop_ts)
        jsonData['topic_counts'] = self._topic_counts
        jsonData['group_counts'] = self._group_counts
        jsonData['rates'] = self._last_calculated_rates

        return jsonData

    @property
    def name(self):
        return self._name

    def rename(self, new_name):
        self._name = new_name

    def topics(self, group):
        if group in self.groups:
            return self._topic_counts[group].keys()
        else:
            return []

    @property
    def groups(self):
        return self._group_counts.keys()

    @property
    def start_ts(self):
        return self._start_ts

    @property
    def uptime(self):
        return round(self._stop_ts - self._start_ts, INTERVAL_PRECISION)


class CounterEvent():
    '''
    This class is for counting events. Each event has a topic and a group.
    Count and rates for each event are calculated for each individual topic as well as at the group level.
    Percentage of each topic within a group is also calculated.
    '''
    def __init__(self, name="undefined", start_ts=time.time()):
        self._name = name
        self._original_start_ts = start_ts
        self._start_ts = start_ts
        self._stop_ts = None

        self._topic_counts = {}
        self._group_counts = {}

        self._topic_latest_ts = {}
        self._group_latest_ts = {}

        self._last_calculated_rates = None

    def increment(self, topic, group="default", increment=1, timestamp=None):
        log.debug(f"topic: {topic} group: {group} increment: {increment} timestamp: {timestamp}")
        if not isinstance(topic, (str, int, float, bool)):
            log.error(f"invalid type for key ({topic} to increment)")
            return None
        if not isinstance(group, (str, int, float, bool)):
            log.error(f"invalid type for group ({group} to increment)")
            return None
        if timestamp is None:
            timestamp = time.time()
        if group not in self.groups:
            self._topic_counts[group] = {}
            self._topic_latest_ts[group] = {}
        if topic in self.topics(group):
            self._topic_counts[group][topic] += increment
            self._topic_latest_ts[group][topic] = \
                self._topic_latest_ts[group][topic] if self._topic_latest_ts[group][topic] > timestamp else timestamp
        else:
            self._topic_counts[group][topic] = increment
            self._topic_latest_ts[group][topic] = timestamp

        if group in self.groups:
            self._group_counts[group] += increment
            self._group_latest_ts[group] = self._group_latest_ts[group] \
                if self._group_latest_ts[group] > timestamp else timestamp
        else:
            self._group_counts[group] = increment
            self._group_latest_ts[group] = timestamp

    def _stats(self, stop_ts=time.time(), interval=None):
        '''
        calculates the stats of the counters based on the counts
        from start_ts to stop_ts.  If interval is defined, use the
        value (in seconds) as the duration of the counter rather
        than start_ts and stop_ts
        '''
        self._stop_ts = stop_ts

        if interval:
            interval_for_calculation = interval
        else:
            interval_for_calculation = round(stop_ts - self._start_ts, INTERVAL_PRECISION)

        if interval_for_calculation <= 0:
            log.error(f"Cannot calculate stats for {self.name} counter as the interval is 0")
            return {}

        topic_rates = {}
        group_rates = {}
        topic_percentage = {}

        results = {}
        results['number_of_topics'] = {}

        for group in self.groups:
            results['number_of_topics'][group] = len(self.topics(group))
            for topic, count in self._topic_counts[group].items():
                if group not in topic_rates.keys():
                    topic_rates[group] = {}
                    topic_percentage[group] = {}
                topic_rates[group][topic] = round(count / interval_for_calculation, RATE_PRECISION)
                topic_percentage[group][topic] = round(count / self._group_counts[group], RATE_PRECISION)

        for group, count in self._group_counts.items():
            group_rates[group] = round(count / interval_for_calculation, RATE_PRECISION)

        results['interval'] = interval_for_calculation
        results['number_of_groups'] = len(self.groups)
        results['topic_rates'] = topic_rates
        results['group_rates'] = group_rates
        results['topic_percentage'] = topic_percentage

        return results

    def _update_stats(self, stop_ts=time.time(), interval=None):
        self._last_calculated_rates = self._stats(stop_ts=stop_ts, interval=interval)

    def snapshot(self, stop_ts=None, interval=None, update_stats=False):
        '''returns snapshot in JSON

        Parameters
        ----------
        stop_ts: int (epoch seconds)
            calculate rates based on stop_ts as end timestamp
        interval: int
            not used. backward compatibility
        update_stats: bool
            update stats/rates before returning results
        '''
        if stop_ts is None:
            # Use last recorded stop_ts unless it is not set
            if self._stop_ts is None:
                self._stop_ts = time.time()
            stop_ts = self._stop_ts
        if update_stats:
            self._update_stats(stop_ts=stop_ts)
        self._stop_ts = stop_ts

        jsonData = {}
        jsonData['name'] = self.name
        jsonData['original_start_ts'] = self._start_ts
        jsonData['start_ts'] = self._start_ts
        jsonData['stop_ts'] = self._stop_ts
        jsonData['uptime'] = round(self._stop_ts - self._start_ts, INTERVAL_PRECISION)
        jsonData['original_start_time_str'] = timestamp2str(self._original_start_ts)
        jsonData['start_time_str'] = timestamp2str(self._start_ts)
        jsonData['stop_time_str'] = timestamp2str(self._stop_ts)
        jsonData['topic_counts'] = self._topic_counts
        jsonData['group_counts'] = self._group_counts
        jsonData['topic_latest_ts'] = self._topic_latest_ts
        jsonData['group_latest_ts'] = self._group_latest_ts
        jsonData['rates'] = self._last_calculated_rates

        return DefaultMunch.fromDict(jsonData)

    def snapshot_table(self, stop_ts=None, interval=None, update_stats=False,
                       summary_only=False):
        '''returns snapshot in table format

        Parameters
        ----------
        stop_ts: int (epoch seconds)
            calculate rates based on stop_ts as end timestamp
        interval: int
            not used. backward compatibility
        update_stats: bool
            update stats/rates before returning results
        summary_only: bool
            return only group level stats
        '''
        s = self.snapshot(stop_ts=stop_ts, interval=interval, update_stats=update_stats)

        TABLE_WIDTH = 97
        GROUP_HEADER_FMT = "{:45}    {:>10}    {:>6} {:>14}    {:>8}"
        GROUP_FMT = "{:48} {:>10}     {:>6} {:>14.3f}  {:>8.3f}"
        KEY_FMT = "  {:46} {:10d}     {:6.2f} {:14.3f}  {:8.3f}"
        TS_FMT = "%Y-%m-%d %H:%M:%S"

        stop_ts = s.stop_ts

        start_ts_str = datetime.fromtimestamp(s.start_ts).strftime(TS_FMT)
        stop_ts_str = datetime.fromtimestamp(s.stop_ts).strftime(TS_FMT)
        original_start_ts_str = datetime.fromtimestamp(s.original_start_ts).strftime(TS_FMT)

        o = []
        o.append("=" * TABLE_WIDTH)
        o.append(f"Counter Name: {self.name}      Original Start: {original_start_ts_str}")
        o.append(f"Start: {start_ts_str:>22}     Stop: {stop_ts_str:>22}    Duration: {s.uptime:8.3f} (s)")
        if summary_only:
            o.append(GROUP_HEADER_FMT.format("Group", "Count", "", "Rate (s)", "Age (s)"))
        else:
            o.append(GROUP_HEADER_FMT.format("Group/Key", "Count", "%", "Rate (s)", "Age (s)"))
        o.append("-" * TABLE_WIDTH)
        for group in sorted(s.group_counts.keys()):
            count = s.group_counts[group]
            age = stop_ts - s.group_latest_ts[group]
            o.append(GROUP_FMT.format(f"{group} ({s.rates.number_of_topics[group]})", count, "", s.rates.group_rates[group], age))
            if not summary_only:
                for key in s.topic_counts[group].keys():
                    count = s.topic_counts[group][key]
                    percentage = s.rates.topic_percentage[group][key] * 100
                    rate = s.rates.topic_rates[group][key]
                    age = stop_ts - s.topic_latest_ts[group][key]
                    key_str = key if key is not None else UNKNOWN_LABEL
                    o.append(KEY_FMT.format(key_str, count, percentage, rate, age))
        o.append("")

        return '\n'.join(o)

    def reset(self, name=None, start_ts=None):
        '''resets counter

        Parameters
        ----------
        name: str
            change the name of the counter
        start_ts: int (epoch time)
            reset the start-time of the counter
        '''
        if name is not None:
            self.rename(name)
        if start_ts is None:
            start_ts = time.time()
        self._start_ts = start_ts
        self._stop_ts = None

        self._topic_counts = {}
        self._group_counts = {}

        self._topic_latest_ts = {}
        self._group_latest_ts = {}

        self._last_calculated_rates = None

    @property
    def name(self):
        return self._name

    def rename(self, new_name):
        self._name = new_name

    def topics(self, group):
        if group in self.groups:
            return self._topic_counts[group].keys()
        else:
            return []

    @property
    def groups(self):
        return self._group_counts.keys()

    @property
    def start_ts(self):
        return self._start_ts

    @property
    def original_start_ts(self):
        return self._original_start_ts


class CounterTime():
    '''
    This class is for counting time periods. Each time period has a topic and a group.
    Total and averages for each time period are calculated for each individual topic as well as at the group level.
    '''
    def __init__(self, name="undefined", start_ts=time.time()):
        self._name = name
        self._start_ts = start_ts
        self._stop_ts = None

        self._topic_counts = {}
        self._topic_times = {}
        self._group_counts = {}
        self._group_times = {}

        self._last_calculated_stats = None

    def increment(self, topic, group="default", increment=0.0):
        log.debug(f"topic: {topic} group: {group} increment: {increment}")
        if not isinstance(topic, (str, int, float, bool)):
            log.error(f"invalid type for key ({topic} to increment)")
            return None
        if not isinstance(group, (str, int, float, bool)):
            log.error(f"invalid type for group ({group} to increment)")
            return None
        if group not in self.groups:
            self._topic_counts[group] = {}
            self._topic_times[group] = {}
        if topic in self.topics(group):
            self._topic_counts[group][topic] += 1
            self._topic_times[group][topic] += increment
        else:
            self._topic_counts[group][topic] = 1
            self._topic_times[group][topic] = increment

        if group in self.groups:
            self._group_counts[group] += 1
            self._group_times[group] += increment
        else:
            self._group_counts[group] = 1
            self._group_times[group] = increment

    def _stats(self, stop_ts=time.time(), interval=None):
        '''
        calculates the stats of the counters based on the counts
        from start_ts to stop_ts.  If interval is defined, use the
        value (in seconds) as the duration of the counter rather
        than start_ts and stop_ts
        '''
        self._stop_ts = stop_ts

        if interval:
            interval_for_calculation = interval
        else:
            interval_for_calculation = round(stop_ts - self._start_ts, INTERVAL_PRECISION)

        if interval_for_calculation <= 0:
            log.error(f"Cannot calculate stats for {self.name} counter as the interval is 0")
            return {}

        topic_averages = {}
        group_averages = {}

        results = {}

        for group in self.groups:
            for topic, times in self._topic_times[group].items():
                if group not in topic_averages.keys():
                    topic_averages[group] = {}
                topic_averages[group][topic] = round(float(times) / self._topic_counts[group][topic], 5)

        for group, times in self._group_times.items():
            group_averages[group] = round(float(times) / self._group_counts[group], 5)

        results['interval'] = interval_for_calculation
        results['topic_averages'] = topic_averages
        results['group_averages'] = group_averages

        return results

    def _update_stats(self, stop_ts=time.time(), interval=None):
        self._last_calculated_stats = self._stats(stop_ts=stop_ts, interval=interval)

    def snapshot(self, stop_ts=None, interval=None, update_stats=False):
        if stop_ts is None:
            # Use last recorded stop_ts unless it is not set
            if self._stop_ts is None:
                self._stop_ts = time.time()
            stop_ts = self._stop_ts
        if update_stats:
            self._update_stats(stop_ts=stop_ts)
        self._stop_ts = stop_ts

        jsonData = {}
        jsonData['name'] = self.name
        jsonData['start_ts'] = self._start_ts
        jsonData['stop_ts'] = self._stop_ts
        jsonData['uptime'] = round(self._stop_ts - self._start_ts, INTERVAL_PRECISION)
        jsonData['start_time_str'] = timestamp2str(self._start_ts)
        jsonData['stop_time_str'] = timestamp2str(self._stop_ts)
        jsonData['topic_times'] = self._topic_times
        jsonData['topic_counts'] = self._topic_counts
        jsonData['group_times'] = self._group_times
        jsonData['group_counts'] = self._group_counts
        jsonData['averages'] = self._last_calculated_stats

        return jsonData

    @property
    def name(self):
        return self._name

    def rename(self, new_name):
        self._name = new_name

    def topics(self, group):
        if group in self.groups:
            return self._topic_counts[group].keys()
        else:
            return []

    @property
    def groups(self):
        return self._group_counts.keys()

    @property
    def start_ts(self):
        return self._start_ts


class CounterTrio():
    '''
    This class combines three counters in one: overall, current interval and previous interval.
    counter_clazz MUST be one of CounterEvent or CounterTime.
    '''
    def __init__(self, name="undefined", start_ts=time.time(), counter_clazz=None,
                 overall_name="overall", current_interval_name="current_interval", previous_interval_name="previous_interval"):
        self.counter_clazz = counter_clazz
        self.overall_name = overall_name
        self.overall_counter = counter_clazz(name=overall_name, start_ts=start_ts)
        self.current_interval_name = current_interval_name
        self.current_interval_counter = counter_clazz(name=current_interval_name, start_ts=start_ts)
        self.previous_interval_name = previous_interval_name
        self.previous_interval_counter = None

    def increment(self, topic, group="default", increment=None):
        log.debug(f"topic: {topic} group: {group} increment: {increment}")
        if not isinstance(topic, (str, int, float, bool)):
            log.error(f"invalid type for key ({topic} to increment)")
            return None
        if not isinstance(group, (str, int, float, bool)):
            log.error(f"invalid type for group ({group} to increment)")
            return None
        self.overall_counter.increment(topic, group=group, increment=increment)
        self.current_interval_counter.increment(topic, group=group, increment=increment)

    def snapshot(self, stop_ts=None, interval=None, update_stats=False):
        # returns a JSON object of the current stats
        jsonData = {}
        if stop_ts is None:
            stop_ts = time.time()
        jsonData[self.overall_name] = self.overall_counter.snapshot(stop_ts=stop_ts, update_stats=update_stats)
        jsonData[self.current_interval_name] = self.current_interval_counter.snapshot(stop_ts=stop_ts, update_stats=update_stats)
        if self.previous_interval_counter:
            # Don't update stats again as the data is no longer updated
            jsonData[self.previous_interval_name] = self.previous_interval_counter.snapshot(stop_ts=None, update_stats=False)
        return jsonData

    def rotate_counters(self):
        current_ts = time.time()
        self.previous_interval_counter = self.current_interval_counter
        self.previous_interval_counter.rename(self.previous_interval_name)
        # Update stats one last time as counter retires
        self.previous_interval_counter.snapshot(stop_ts=current_ts, update_stats=True)
        self.current_interval_counter = self.counter_clazz(name=self.current_interval_name, start_ts=current_ts)

    @property
    def name(self):
        return self._name

    def rename(self, new_name):
        self._name = new_name

    def topics(self, group):
        raise Exception("unimplemented")

    @property
    def groups(self):
        raise Exception("unimplemented")



#
# Copyright 2020 Thomas Bastian, Jeffrey Goff, Albert Pang
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
#
# Helpers for parsing table data out of text.
#
# Authors: Thomas Bastian
#
'''

'''
# Parameters
'''
'''
#
'''

import logging
from .RowColumnTableParser import RowColumnTableParser


# Logger
log = logging.getLogger(__name__)


'''
All parsing is configured via table options.

TABLE OPTIONS
-------------
Dictionary that contains common table options and parser specific table options.
The table parser MUST always be specified.

{
   "parser": "aos-std|aos-dp-std|generic-regex|generic-json-csv"
   [, "begin": "<begin>" ]
   [, "data.trim": false|true ]
   [, "end": "<end>" ]
   [, "json.lines": "<key>" ]
   [, "marker": "<marker>" ]
   [, <parser specific table options> ]
}

TABLE PARSERS
-------------
Following table parsers are available (key 'parser'):
- "aos-std": understands standard AOS table output such as "show ap active",
             "show user", "show ap bss-table", etc...
- "aos-dp-std": understands standard AOS datapath table output such as
                "show datapath tunnel table", etc...
- "generic-regex": generic regular expression based table parser.
- "generic-json-csv": generic JSON CSV like output table parser.

COMMOM TABLE OPTIONS
--------------------
All parsers will honor following optional parameters:
- "begin": "<begin regex>". Defines a regular expression that will mark the
           beginning of the CLI output to use for parsing the table. This will
           typically be needed when the CLI output holds several tables.
           Default: None
           NOTE: when the provided <begin regex> does not match, the result will
                 be like no <begin regex> was provided.
- "end": "<end regex>". Defines a regular expression that will mark the end
         of the CLI output to use for parsing the table. This will typically be
         needed when the table has trailing output that is not part of the table.
         Default: None
         NOTE: when the provided <end regex> does not match, the result will
               be like no <end regex> was provided.
               The provided <end regex> applies to CLI output AFTER <begin regex>
               has matched or if no <begin regex> was provided to beginning of CLI
               output.
- "marker": "<marker>". Provide table marker, such as "Active AP Table" for table "show ap active".
            Default None, which auto-detects the marker based on the first lines of output.
- "json.lines": "<key>". By default input is assumed to be text. When "json.lines" is specified, input is expected
                to be a JSON object where the "json.lines" provided key specifies an array of text lines, such as:
                { "_data": [ "line1", "line2", "line3" ] } (i.e. "json.lines": "_data")
                Default: None
- "data.trim": true|false. When true, all table values have beginning and ending white-spaces trimmed.
               Default: true

aos-std TABLE PARSER OPTIONS
----------------------------
None in addition to the common table parser parameters.

aos-dp-std TABLE PARSER OPTIONS
-------------------------------
None in addition to the common table parser parameters.

generic-regex TABLE PARSER OPTIONS
----------------------------------
The generic regular expression based parser supports three broad use cases:
- header driven split-up based on spacing between column names, specify:
  "header.offset" and "data.offset" parameters and optionally "stop.on.empty.line".
- header driven split-up based on regex with named groups, specify:
  "header.offset", "header.regex" and "data.offset" parameters and optionally
  "stop.on.empty.line".
- data driven split-up based on regex with named groups, specify:
  "data.offset" and "data.regex" parameters and optionally "stop.on.empty.line" and/or
  "skip.partial.match".

Parameter details:
NOTE: all offsets are relative to the FIRST non blank line in CLI output, i.e. offset
      0 is the first non blank line.
"header.offset": sets the relative offset of the header line.
"header.regex": sets the regular expression to be used to break-up the header line.
"data.offset": sets the relative offset of the first data line
"data.regex": sets the regular expression to be used to break-up each data line.
"stop.on.empty.line": true or false, defaults to true. When true, parsers stops on first
                      blank/empty line.
"skip.partial.match": true or false, defaults to false. When true, parser ignores data lines
                      that don't match the data regular expression.

generic-json-csv TABLE PARSER OPTIONS
-------------------------------------
None in addition to the common table parser parameters.

'''


'''
#
# Parse content as list of dictionaries
#
'''
def contentAsListOfDict(content, marker=None, headerList=None, tableOptions=None):
    if tableOptions is not None:
        marker = tableOptions.get('marker', marker)
    if tableOptions is not None and 'json.lines' in tableOptions:
        # Lines are already in an array
        source = content[tableOptions['json.lines']]
    else:
        # Assume text
        source = content
    parser = RowColumnTableParser(source, marker=marker, tableOptions=tableOptions)
    parser.process()
    parsedTable = parser.getTables()[0]
    # Data trim, default true
    if (tableOptions is None) or ('data.trim' in tableOptions and tableOptions['data.trim']) or ('data.trim' not in tableOptions):
        # Header
        for header in parsedTable['header']:
            header['title'] = header['title'].strip()
        # Data
        for row in parsedTable['rows']:
            i = 0
            for value in row:
                row[i] = value.strip()
                i += 1
    return tableAsListOfDict(parsedTable, headerList=headerList)


'''
#
# Parse content as parameter value list
#
'''
def contentAsParameterValueList(content, marker=None):
    source = content
    parser = RowColumnTableParser(source, marker=marker)
    parser.process()
    parsedTable = parser.getTables()[0]
    # Data
    parameterValueList = dict()
    for row in parsedTable['rows']:
        parameterValueList[row[0].strip()] = row[1].strip()
    return parameterValueList


'''
#
# Table as list of dictionaries
#
'''
def tableAsListOfDict(table, headerList=None):
    # Fill list
    trows = table['rows']
    theader = table['header']
    valueList = []
    if headerList is not None:
        for cheader in theader:
            headerList.append(cheader['title'])
    for row in trows:
        rowd = dict()
        for c in range(0, len(row)):
            rowd[theader[c]['title']] = row[c]
        valueList.append(rowd)
    return valueList



#
# Copyright 2021 Thomas Bastian, Jeffrey Goff, Albert Pang
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
#
# CliParser base classes
#
# Authors: Thomas Bastian
#
'''

'''
# Parameters
'''
'''
#
'''

import logging
from collections import UserDict


# Logger
log = logging.getLogger(__name__)


# Generic Parser Request
class CliParserRequest(UserDict):
    def __init__(self):
        super().__init__()

    # Hook to clone
    def clone(self):
        c = self.__class__()
        c.update(self)
        return c


# Generic interface for all parsers
class CliParser():
    def __init__(self):
        pass

    def parse(self, parserRequest):
        return None



#
# Copyright 2020 Thomas Bastian, Jeffrey Goff, Albert Pang
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
#
# Implements a generic table CLI output parser for JSON CSV like output.
# (a JSON array with a header row followed by data rows, each row encoded as a JSON array)
#
# Authors: Thomas Bastian
#
'''

'''
# Parameters
'''
'''
#
'''

import json
import logging
from .AbstractTableParser import AbstractTableParser


# Logger
log = logging.getLogger(__name__)


class GenericJsonCsvTableParser(AbstractTableParser):

    def __init__(self, source, tableOptions={}):
        super().__init__()
        self.source = source

    def readContent(self):
        # TODO: check if jsonData is an array
        self.tableData = json.loads(self.source)

    def parse(self):
        tsize = len(self.tableData)
        log.debug("json table size: [%d]" % (tsize))
        if tsize <= 1:
            # No data
            return

        # Process header
        theader = self.tableData[0]
        hsize = len(theader)
        log.debug("json table header size: [%d]" % (hsize))
        if hsize <= 0:
            # No columns
            return
        columns = []
        for hdata in theader:
            column = { 'title': hdata, 'begin': -1, 'end': -1 }
            columns.append(column)
            log.debug("json table header column: [%s]" % (column['title']))

        # Main table
        self.getTables()[0]['name'] = "generic json csv table"
        rows = []
        self.tables[0]['rows'] = rows
        self.tables[0]['header'] = columns

        # Process rows
        for i in range(1, tsize):
            trow = self.tableData[i]
            rows.append(trow)
            log.debug("json table data row: [%d] [%s]" % (i, str(trow)))
